import sys,os

sys.path.append(os.path.join(os.path.dirname(__file__),'../external'))
sys.path.append(os.path.join(os.path.dirname(__file__),'../external/benchmark_database'))
sys.path.append(os.path.join(os.path.dirname(__file__),'../external/com_github_interaction_dataset_interaction_dataset/python'))

